from ezfsm.TimingSeriesFunction import DataArea
from ezfsm.vessel import Vessel
class UnsupportedTransFuncs(Exception):
    def __init__(self, ss):
        super(UnsupportedTransFuncs, self).__init__("\n\n[Error]:TransFuncs with type<%s> is not a valid transFuncs." % str(type(ss)))


class TransFuncKwargs(object):
    key = None
    def __init__(self, value):
        self.value = value

    def __call__(self, *args, **kwargs):
        return {self.key: self.value}

class IfDoc(TransFuncKwargs):
    key = 'cond_doc'
CondDoc = IfDoc

class DoDoc(TransFuncKwargs):
    key = 'exec_doc'
ExecDoc = DoDoc

class BufTs(TransFuncKwargs):
    key = 'buffer_times'
BufTimes = BufTs
BufferTimes = BufTs

class BufFn(TransFuncKwargs):
    key = 'buffer_execution'
BufDo = BufFn
BufFunc = BufFn
BufExec = BufFn
BufExecution = BufFn
BufferFunc = BufFn
BufferExec = BufFn
BufferExecution = BufFn

class BufDoc(TransFuncKwargs):
    key = 'buffer_doc'
BufferDoc = BufDoc

class ActionBuffer(Vessel):
    def __init__(self, max_times, execution, doc, auto_reset=False):
        super(ActionBuffer, self).__init__(max_times)
        self.exec, self.doc = execution, doc
        self.__ar = auto_reset

    def Use(self):
        self.DoDelta(-1)
        if self.exec:
            self.exec()
        if self.IsEmpty():
            if self.__ar:
                self.Reset()
            return True


    def Reset(self):
        self.Full()


class TransFunc(object):  # 最全面的转移函数
    inst = None
    def __init__(self, srt, dst, trans_condition, trans_execution, priority=0, cond_doc="", exec_doc="", buffer_times=0, buffer_execution=None, buffer_doc=''):
        """

        :param srt: 起点的name, 可以是多个节点
        :param dst: 终点的name, 可以是多个节点
        所有节点间两两相连

        :param trans_condition: 触发该转换的条件, 是一个函数 ()，返回True或False
        :param trans_execution: 转换发生时执行的代码. 是一个函数 ()
        :param priority: 检查优先级，越高越先检查
        :param cond_doc: 允许你添加条件说明文档，用于用户分析流程图
        :param exec_doc: 允许你添加转移说明文档，用于用户分析流程图
        :param buffer_times: 允许你设置进入此转移函数所需的缓冲次数
        缓冲:
            当IF的条件检查通过时，状态机会试图执行转移操作
            但是若是存在n次缓冲，则状态机在使n减少1。且在n <= 0前都不执行转移操作
            以下操作会使n被重置:
                1.n次缓冲完毕、执行转移操作后
                2.该state的另一个分支的转移操作被执行
                3.被StateMachine.ResetBuffer重置.
        :param buffer_execution: 触发buffer后执行的代码
        :param buffer_doc: 允许你添加缓冲函数说明文档，用于用户分析流程图
        """
        assert isinstance(srt, (str, tuple, list)), "[Error]:srt can only be str、tuple or list."
        assert isinstance(dst, (str, tuple, list)), "[Error]:srt can only be str、tuple or list."
        assert isinstance(priority, (int, float)), "[Error]:priority can only be int or float."
        self.srt = [srt] if isinstance(srt, str) else list(srt)
        self.dst = [dst] if isinstance(dst, str) else list(dst)
        self.priority = priority
        self.cond = trans_condition
        self.exec = trans_execution
        self.cond_doc = cond_doc
        self.exec_doc = exec_doc
        self.buffer = ActionBuffer(buffer_times, buffer_execution, buffer_doc)

        self.__args, self.__kwds, self.this = (), {}, DataArea()

    def Enable(self, inst):
        self.inst = inst

    def SetPass(self, *args, **kwargs):
        self.__args, self.__kwds = args, kwargs

    def Trig(self):
        if self.cond(*self.__args, **self.__kwds):
            if self.buffer.Use():
                self.exec(*self.__args, **self.__kwds)
                return True
        return False

    def __call__(self, *args, **kwargs):
        self.Trig()

    def __str__(self):
        return "<TransFunc> <%d> %s -> %s\n\tcond: %s\n\texec: %s\n%s"%(
            self.priority,
            self.srt,
            self.dst,
            self.cond,
            self.exec,
            "" if self.buffer.Max() == 0 else "\tbuffer: %d\n\tbuf_exec: %s"%(self.buffer.Max(), self.buffer.exec)
        )

    trig = Trig



class UnsupportedStateSet(Exception):
    def __init__(self, ss):
        super(UnsupportedStateSet, self).__init__("\n\n[Error]:State sets with type<%s> is not a valid State-Set." % str(type(ss)))


class StateSet(object):  # 状态集
    inst = None

    def __init__(self, names=()):
        self.sets = list(names)
        self.state = None

    def Enable(self, inst):
        self.inst = inst

    def Add(self, name, ignore=False):
        if name in self.sets:
            if ignore: return
            raise Exception("[Error]: 重复的状态<{}>.".format(name))
        self.sets += [name]

    def Remove(self, name, ignore=False):
        if name not in self.sets:
            if ignore: return
            raise Exception("[Error]: 待删除的状态<{}>不存在.".format(name))
        self.sets.remove(name)

    def __iter__(self):
        return iter(self.sets)

    def __len__(self):
        return len(self.sets)

    def __getattr__(self, item):
        return getattr(self.sets, item)

    def __getitem__(self, item):
        return self.sets[item]

    def __setitem__(self, key, value):
        self.sets[key] = value

    Del = Remove
    add = Add
    remove = Remove



from ezfsm import *    # This step is best to import all

tfs = [
    ["a", "b", lambda: True, lambda: print("a -> b")]
]

sm = StateMachine(tfs)    # You can also use SM instead of Statemachine

sm.Build()    # compile


#  Ususally, other general FSM does not need to be compiled
#  But this step is set to standardize the data format and improve efficiency. It can make writing more convenient
#  Statemachine without compilation cannot execute most operations.


while not sm.IsFinish(): sm.Execute()    # You can also use sm() instead of sm.Execute()

# if you want to view the graph, you must pip install graphviz and download the graphviz software.

graph = sm.StateGraph()    # get the graphviz.Digraph object
graph.view()



class DataArea(object):
    def iadd(self, attr, value):
        new = getattr(self, attr) + value
        setattr(self, attr, new)
        return new

    def isub(self, attr, value):
        new = getattr(self, attr) - value
        setattr(self, attr, new)
        return new

    def set(self, attr, value):
        setattr(self, attr, value)





"""
允许为用户的函数提供额外的可供访问的运行环境:
    mach: 当前正在运行的状态机 的运行环境
    inst: 当前状态节点 的运行环境
    dest: 目标状态节点 的运行环境
    this: 转移函数自身 的运行环境

    性能测试:(测试用函数见尾部)(单位s)
    传统1: 0.10
    委托1: 0.29
    传统2: 0.16
    委托2: 3.81
    委托2(快速模式): 1.9
    委托2(单次全局设置): 0.34

"""
from ezfsm.DataArea import *

_base_machine, _base_inst, _base_dest, _base_this = DataArea(), DataArea(), DataArea(), DataArea()

_proxy_binds = {}

class Proxy:
    def __init__(self):
        _proxy_binds[self] = None

    def _proxy_bind(self, obj):
        _proxy_binds[self] = obj

    def __getattr__(self, item):
        return getattr(_proxy_binds[self], item)

    def __setattr__(self, key, value):
        setattr(_proxy_binds[self], key, value)


mach, inst, dest, this = Proxy(), Proxy(), Proxy(), Proxy()
mach._proxy_bind(_base_machine)
inst._proxy_bind(_base_inst)
dest._proxy_bind(_base_dest)
this._proxy_bind(_base_this)
GlobalArea = DataArea()

def ChangeGlobal(area, _machine: DataArea, _inst: DataArea, _dest: DataArea, _this: DataArea):
    area.rec_machine, area.rec_inst, area.rec_dest, area.rec_this = mach._proxy_obj, inst._proxy_obj, dest._proxy_obj, this._proxy_obj
    mach._proxy_bind(_machine)
    inst._proxy_bind(_inst)
    dest._proxy_bind(_dest)
    this._proxy_bind(_this)

def ChangeGlobal_Machine(area, _machine: DataArea):
    area.rec_machine = _proxy_binds[mach]
    mach._proxy_bind(_machine)

def ChangeGlobal_Inst(area, _inst: DataArea):
    area.rec_inst = _proxy_binds[inst]
    inst._proxy_bind(_inst)

def ChangeGlobal_Dest(area, _dest: DataArea):
    area.rec_dest = _proxy_binds[dest]
    dest._proxy_bind(_dest)

def ChangeGlobal_This(area, _this: DataArea):
    area.rec_this = _proxy_binds[this]
    this._proxy_bind(_this)

def ChangeGlobalBack(area):
    if getattr(area, 'rec_machine') and getattr(area, 'rec_inst') and getattr(area, 'rec_dest') and getattr(area, 'rec_this'):
        mach._proxy_bind(area.rec_machine)
        inst._proxy_bind(area.rec_inst)
        dest._proxy_bind(area.rec_dest)
        this._proxy_bind(area.rec_this)
    else:
        raise Exception("\n\n[Error]: This DataArea Doesn't contain needy attributes.")

def ChangeGlobalBack_NoCheck(area):
    mach._proxy_bind(area.rec_machine)
    inst._proxy_bind(area.rec_inst)
    dest._proxy_bind(area.rec_dest)
    this._proxy_bind(area.rec_this)

def OuterAreaCall(func, _machine: DataArea, _inst: DataArea, _dest: DataArea, _this: DataArea, *args, **kwargs):
    """
    允许在运行特定代码期间修改machine、 inst、 this, 使其重新临时委托绑定到新的对象
    :param func: 待调用的函数
    :param _machine: 运行期间特定的machine
    :param _inst: 运行期间特定的inst
    :param _dest: 运行期间特定的dest
    :param _this: 运行期间特定的this
    :param args: 待调用的函数(func) 的*args参数
    :param kwargs: 待调用的函数(func) 的**kwargs参数
    :return: 待调用的函数(func) 的返回值
    """
    area = DataArea()
    ChangeGlobal(area, _machine, _inst, _dest, _this)  # DataArea doesn't need some attrs in the first step
    _return = func(*args, **kwargs)
    ChangeGlobalBack_NoCheck(area)
    return _return


def GlobalOuterAreaCall(func, _machine: DataArea, _inst: DataArea, _dest: DataArea, _this: DataArea, *args, **kwargs):
    """
    使用全局GlobalArea，其余同上
    """
    ChangeGlobal(GlobalArea, _machine, _inst, _dest, _this)  # DataArea doesn't need some attrs in the first step
    _return = func(*args, **kwargs)
    ChangeGlobalBack_NoCheck(GlobalArea)
    return _return


def QuickOuterAreaCall(func, _machine: DataArea, _inst: DataArea, _dest: DataArea, _this: DataArea, *args, **kwargs):
    """
    允许在运行特定代码期间修改machine、 inst、 this, 使其重新委托绑定到新的对象
    :param func: 待调用的函数
    :param _machine: 运行期间特定的machine
    :param _inst: 运行期间特定的inst
    :param _dest: 运行期间特定的dest
    :param _this: 运行期间特定的this
    :param args: 待调用的函数(func) 的*args参数
    :param kwargs: 待调用的函数(func) 的**kwargs参数
    :return: 待调用的函数(func) 的返回值
    """
    mach._proxy_bind(_machine)
    inst._proxy_bind(_inst)
    dest._proxy_bind(_dest)
    this._proxy_bind(_this)
    _return = func(*args, **kwargs)
    return _return



"""
from TimingSeriesFunction import *
import time

my_this = DataArea()


def test():
    this.a = 10


def test2():
    my_this.a = 10


this.a = 999
my_this.a = 114.514

a = time.time()
for i in range(921600):
    my_this.a = 10
print("传统赋值耗时: {}s".format(time.time() - a))

a = time.time()
for i in range(921600):
    this.a = 10
print("包装赋值耗时: {}s".format(time.time() - a))

a = time.time()
for i in range(921600):
    test2()
print("传统赋值耗时(函数): {}s".format(time.time() - a))

a = time.time()
for i in range(921600):
    OuterAreaCall(test, mach, inst, my_this)
print("包装赋值耗时(函数): {}s".format(time.time() - a))

a = time.time()
for i in range(921600):
    QuickOuterAreaCall(test, mach, inst, my_this)
print("包装赋值耗时(函数)(快速): {}s".format(time.time() - a))

a = time.time()
ChangeGlobal(GlobalArea, mach, inst, my_this)
for i in range(921600):
    test()
ChangeGlobalBack(GlobalArea)
print("包装赋值耗时(函数)(单次): {}s".format(time.time() - a))
exit()
"""



from ezfsm.StateMachine import *

__version__ = "0.2.0"
__author__ = "MOS.Eagle's Baby"
__time__ = "2021 11 10"
__software__ = "Pycharm"

"""
0.0.1: basic functions
0.1.0: basic function stable version
0.2.0: support for micropython
"""

"""
Example: <Basic>
from ezfsm import *    # This step is best to import all

tfs = [
    ["a", "b", lambda: True, lambda: print("a -> b")]
]

sm = StateMachine(tfs)    # You can also use SM instead of Statemachine

sm.Build()    # compile


#  Ususally, other general FSM does not need to be compiled
#  But this step is set to standardize the data format and improve efficiency. It can make writing more convenient
#  Statemachine without compilation cannot execute most operations.


while not sm.IsFinish(): sm.Execute()    # You can also use sm() instead of sm.Execute()

# if you want to view the graph, you must pip install graphviz and download the graphviz software. 

graph = sm.StateGraph()    # get the graphviz.Digraph object
graph.view()


# more examples on url https://gitee.com/eagle-s_baby/fsm
"""



class Vessel(object):
    def __init__(self, *args):
        self.__min, self.__current, self.__max = 0, 100, 100
        self.__delta_cb, self.__full_cb, self.__empty_cb = None, None, None
        _len = len(args)
        if _len == 1:  # max
            self.RawSetValues(0, args[0], args[0])
        elif _len == 2:  # min max
            self.RawSetValues(args[0], args[1], args[1])
        elif _len == 3:  # min, current, max
            self.RawSetValues(*args)

    def RawSetValues(self, min, current, max):
        """
        不会触发回调函数, 不会触发检测
        :param min: double
        :param current: double
        :param max: double
        :return:
        """
        self.__min, self.__current, self.__max = min, current, max

    def SetWhenEmpty(self, empty_func: callable):
        """
        当容器为空时执行的操作
        :param empty_func: 形参: empty_func(vessel), None表示不作任何操作
        :return:
        """
        self.__empty_cb = empty_func

    def SetWhenFull(self, full_func: callable):
        """
        当容器为满时执行的操作
        :param full_func: 形参: full_func(vessel), None表示不作任何操作
        :return:
        """
        self.__full_cb = full_func

    def SetWhenDelta(self, delta_func: callable):
        """
        当容器余量改变时执行的操作
        :param delta_func: 形参: delta_func(vessel, delta:int|float), None表示不作任何操作
        :return:
        """
        self.__delta_cb = delta_func

    def DoDelta(self, delta):
        """
        更新current的值，并依据最后情况调用回调函数
        :param delta: 变化值
        :return:
        """
        last = self.__current
        self.__current += delta
        if self.__max < self.__current:
            self.__current = self.__max
        if self.__current < self.__min:
            self.__current = self.__min

        if self.__delta_cb:
            delta = self.__current - last
            if abs(delta) > 1e-8:
                self.__delta_cb(self, delta)
        if self.__full_cb and self.__max == self.__current:
            self.__full_cb(self)
        if self.__empty_cb and self.__min == self.__current:
            self.__empty_cb(self)

    def Max(self, max = None):
        """
        修改或获取max
        :param max:
        :return:
        """
        if max is not None:
            self.__max = max
            if self.__max < self.__min:
                self.__min = self.__max
            if self.__current > self.__max:
                DoDelta(self.__max - self.__current)
        else:
            return self.__max

    def Min(self, min = None):
        """
        修改或获取min
        :param min:
        :return:
        """
        if min is not None:
            self.__min = min
            if self.__max < self.__min:
                self.__max = self.__min
            if self.__current < self.__min:
                DoDelta(self.__min - self.__current)
        else:
            return self.__min

    def Now(self, now = None):
        """
        修改或获取current
        :param now:
        :return:
        """
        if now is not None:
            DoDelta(now - self.__current)
        else:
            return self.__current

    def Percent(self, per = None):
        """
        修改或获取percent
        :param per:
        :return:
        """
        if per is not None:
            self.DoDelta(self.__min + (self.__max - self.__min) * per - self.__current)
        else:
            return (self.__current - self.__min) / (self.__max - self.__min)

    def IsEmpty(self):
        return self.__min >= self.__current  # 一般情况下min不会小于current， 但是不排除用户使用RawSet的情况

    def IsFull(self):
        return self.__max <= self.__current

    def Full(self):
        self.DoDelta(self.__max - self.__current)

    def Empty(self):
        self.DoDelta(self.__min - self.__current)

    def __add__(self, other):
        return self.__current + other

    def __iadd__(self, other):
        self.DoDelta(other)

    def __sub__(self, other):
        return self.__current - other

    def __isub__(self, other):
        self.DoDelta(-other)

    def __int__(self):
        return int(self.__current)

    def __float__(self):
        return float(self.__current)

    def __str__(self):
        return """Vessel{
            .min: {}
            .current: {}
            .max: {}
            percent: {}
        }
        """.foramt(self.__min, self.__current, self.__max, self.Percent())



from ezfsm.TimingSeriesFunction import *
from ezfsm.TransFunction import *
from ezfsm.StateNode import *
from ezfsm.StateSet import *

class StateNotInSets(Exception):
    def __init__(self, name):
        super(StateNotInSets, self).__init__("\n\n[Error]: state <{}> doesn't exist in state sets.".format(name))

class NotImplementBuildUp(Exception):
    def __init__(self):
        super(NotImplementBuildUp, self).__init__("\n\n[Error]: StateMachine need build first.")

class StateMachine_const(object):
    RAW = 0
    READY = 1
    BUSY = 2
    FINISH = 3


class StateMachine_cb(StateMachine_const):
    end_cb = None
    trans_cb = None
    start_cb = None
    def SetWhenEnd(self, end_func):
        """
        设置结束回调
        :param end_func: end_func(self), None表示关闭回调
        :return:
        """
        self.end_cb = end_func

    def SetWhenTrans(self, trans_func):
        """
        设置转移结束回调
        :param trans_func: trans_func(self, srt, dest, tf), None表示关闭回调
        :return:
        """
        self.trans_cb = trans_func

    def SetWhenStart(self, start_func):
        """
        设置启动回调
        :param start_func: start_func(self), None表示关闭回调
        :return:
        """
        self.start_cb = start_func

class StateMachine_area(StateMachine_const):
    machine_area = DataArea()
    inst_areas = {}  # 编译前不可使用
    this_areas = {}  # 编译前不可使用
    ready = 0
    st = {}
    def ResetAreas(self):
        if self.ready < self.READY:
            raise NotImplementBuildUp()
        self.machine_area = DataArea()
        self.inst_areas = {}  # 编译前不可使用
        self.this_areas = {}  # 编译前不可使用

        for k in self.st:
            self.inst_areas[k] = DataArea()
            self.this_areas[k] = []
            for item in self.st[k]:
                self.this_areas[k] += [DataArea()]


class StateMachine_data(object):
    NOT = 0
    MAYBE = 1
    SURE = 2
    def __init__(self, *args, start=None, end=None):
        """
        用户可以只传trans_funcs这个参数，PyState会自动提取节点名
        :param arg1: 可以为state sets (记做sets tfs 正常模式)， 也可以是tfs (记做auto-sets tfs 自动补全模式)
        :param arg2: 正常模式时: tfs; 自动补全模式时: start(可省略)
        :param arg3: 正常模式时: start(可省略); 自动补全模式时: end(可省略)
        :param arg4: 正常模式时: end(可省略); 自动补全模式时: 不需要该参数，传入会产生歧义，可能会导致程序产生Exception

        当start和end通过kwargs传递时，可以忽略args中的start和end
        """
        self.raw_ss, self.raw_tfs = None, None
        self.start, self.end, self.now = None, None, None

        self.ParseArgs(args)
        if start:
            self.start = start
        if end:
            self.end = end
        self.dis_rec_log_flag = True
        self.show_log_flag = True
        self.raw_ss_auto_fill_flag = False


        self.log = []

    def ParseArgs(self, args):
        """
        解析用户的参数，判断类别
        :param args:
        :return:
        """
        _len = len(args)
        if not _len:  # _len == 0:
            raise Exception("\n\n[Error]: Get nothing from the params. ")
        elif _len == 1:  # tfs mode
            self.raw_ss, self.raw_tfs = None, args[0]
        elif _len == 2:  # sets-tfs mode | tfs mode + start
            if self.CanBeTransFuncs(args[0]):
                self.raw_ss, self.raw_tfs, self.start = None, args[0], args[1]
            elif self.CanBeStateSets(args[0]):
                self.raw_ss, self.raw_tfs = args
            else:
                raise Exception("\n\n[Error]: Unsupported Pass params. Please check your params.")
        elif _len == 3:  # sets-tfs mode + start | tfs mode + start + end
            if self.CanBeTransFuncs(args[0]):
                self.raw_ss, self.raw_tfs, self.start, self.end = None, args[0], args[1], args[2]
            elif self.CanBeStateSets(args[0]):
                self.raw_ss, self.raw_tfs, self.start = args
            else:
                raise Exception("\n\n[Error]: Unsupported Pass params. Please check your params.")
        elif _len == 4:  # sets-tfs mode + start + end
            self.raw_ss, self.raw_tfs, self.start, self.end = args
        else:
            raise Exception("\n\n[Error]: Expected params up to 4, but got " + str(_len))

    def CanBeStateSets(self, item):
        """
        state sets 只能是非空一维可变可迭代对象

        注意: item必须是可以变动内容值的容器(不可变动值的容器只支持tuple)
        注意: trans funcs能通过此检查
        :param item:
        :return:
        """
        try:
            iter(item)
            if not isinstance(item, tuple):
                item[0] = item[0]
            try:
                iter(item)
                item[0][0] = item[0][0]
                return self.NOT
            except:
                return self.SURE
        except:
            return self.NOT


    def CanBeTransFuncs(self, item):
        """
        trans funcs
            可以是非空一维TF对象， SURE
            可以是非空二维可变可迭代对象， SURE
            可以是字符串对象 MAYBE

        注意: item必须是可以变动内容值的容器(不可变动值的容器只支持tuple)
        注意: state sets不能通过此检查
        :param item:
        :return:
        """
        if isinstance(item, str):
            return self.MAYBE

        try:
            iter(item)
            if not isinstance(item, tuple):
                item[0] = item[0]
            try:
                iter(item[0])
                if not isinstance(item[0], tuple):
                    item[0][0] = item[0][0]
                try:
                    iter(item[0][0])
                    item[0][0][0] = item[0][0][0]
                    return self.NOT
                except:
                    return self.SURE
            except:
                if isinstance(item[0], TransFunc):
                    return self.SURE
                else:
                    return self.NOT

        except:
            return self.NOT

    def TidyStateSets(self, ss):
        """
        标准化用户传入的state sets
        因为这些ss可能不是标准的状态集
        在BuildUp前都不做转化，允许用户修改它们的源参数
        :param ss:
        :return:
        """
        if isinstance(ss, StateSet):
            return ss
        elif isinstance(ss, (tuple, list)):
            return StateSet(ss)
        else:
            try:
                iter(ss)
                print("[Warning]: %s is an unusual iterable object<type: %s>. It may cause Exception in running. "%(str(ss), str(type(ss))))
                self.Log("Trans an unusual iterable object<type: {}> to general StateSet.".format(type(ss)))
                return StateSet(list(ss))
            except:
                raise UnsupportedStateSet(ss)


    def TidyTransFuncs(self, tfs):
        """
        标准化用户传入的trans funcs
        因为这些tfs可能不是标准的转移集
        在BuildUp前都不做转化，允许用户修改它们的源参数
        :param tfs:
        :return:
        """
        if isinstance(tfs, str):
            return self.ParseStringTransFuncs(tfs)
        else:
            try:
                iter(tfs)
                if not tfs: raise Exception("\n\n[Error]: transFuncs can not be empty.")
                if isinstance(tfs[0], TransFunc):
                    return list(tfs)
                try:
                    iter(tfs[0])
                    # 在初始化时已经检查过了，不应该是三维的
                    return self.Parse2DTransFuncs(tfs)
                except:
                    raise UnsupportedTransFuncs(tfs)
            except:
                raise UnsupportedTransFuncs(tfs)

    @staticmethod
    def Parse2DTransFuncs(params_tfs):
        try:
            tfs = []
            for _args in params_tfs:
                args, kwargs = [], {}
                for arg in _args:
                    if isinstance(arg, TransFuncKwargs):
                        kwargs.update(arg())
                    else:
                        args += [arg]

                tfs += [TransFunc(*args, **kwargs)]
            return tfs
        except:
            raise Exception("\n\n[Error]: Try parse 2DTransFunc, but failed. Please check your params.")

    def ParseStringTransFuncs(self, str_tfs):
        """
        将字符串格式的转移描述 转换为 TF描述
        :param str_tfs:
        :return:
        """
        raise Exception("\n\n[Error]:暂未开放. It will be available later.")

    def Log(self, add=None):
        if add is None:
            return self.log
        else:
            if self.dis_rec_log_flag:
                self.log += [add]
                if self.show_log_flag: print(add)

    def EnableLogRecord(self, value=True):
        self.dis_rec_log_flag = value

    def EnableLogPrint(self, value=True):
        self.show_log_flag = value

    def AddState(self, *states):
        self.raw_ss += list(states)
        for state in states:
            self.Log("User add state: <%s>"%state)

    def AddTransFunc(self, *tfs):
        self.raw_tfs += list(tfs)
        for tf in tfs:
            self.Log("User add transFunc: <%s>" % tf)

    def Config(self, start=None, end=None):
        """
        设置起点和终点
        :param start: str, None时不修改此项
        :param end: str, None时不修改此项
        :return:
        """
        if start is not None:
            if self.raw_ss and start not in self.raw_ss:
                raise StateNotInSets(start)
            self.start = start
        if end is not None:
            if self.raw_ss and end not in self.raw_ss:
                raise StateNotInSets(end)
            self.end = end

    def ResetBuffer(self, srt, dst):
        """
        重置指定连接的缓冲
        :param srt:
        :param dst:
        :return:
        """
        if not self.st.get(srt): raise StateNotInSets(srt)
        if not self.st.get(dst): raise StateNotInSets(dst)
        for _srt, _dst, _tf in self.st[srt]:
            if _dst == dst:
                _tf.buffer.Reset()

    def GetTidiedInputParams(self):
        ss, tfs = [], []
        if self.raw_ss is None:
            self.raw_ss_auto_fill_flag = True
        else:
            ss = self.TidyStateSets(self.raw_ss)
        
        if self.raw_tfs is None: 
            raise Exception("\n\n[Error]:TransFuncs can not be empty.")
        else:
            tfs = self.TidyTransFuncs(self.raw_tfs)
        
        return ss, tfs

    def TidyStartAndEnd(self, raw_ss, raw_tfs):
        if self.start is None or self.end is None:
            if self.raw_ss_auto_fill_flag:
                if self.start is None and self.end is None and len(raw_tfs) >= 1:
                    self.start, self.end = raw_tfs[0].srt[0], raw_tfs[-1].dst[-1]
                    self.Log("No set <start> <end>, Auto Set : {start <- %s, end <- %s}" % (self.start, self.end))
                elif self.start is None and self.end is not None and len(raw_tfs) >= 1:
                    self.start = raw_tfs[0].srt[0]
                    self.Log("No set <start>, Auto Set : {start <- %s}" % self.start)
                elif self.start is not None and self.end is None and len(raw_tfs) >= 1:
                    self.end = raw_tfs[-1].dst[-1]
                    self.Log("No set <end>, Auto Set : {end <- %s}" % self.end)
                else:
                    raise Exception(
                        "\n\n[Error]: StateMachine must have a start-node and a end-node.\n\t--At least, you need pass transFuncs which length >= 1 while you don't pass stateSets.")

            else:
                if self.start is None and self.end is None and len(raw_ss) >= 2:
                    self.start, self.end = raw_ss[0], raw_ss[-1]
                    self.Log("No set <start> <end>, Auto Set : {start <- %s, end <- %s}"%(self.start, self.end))
                elif self.start is None and self.end is not None and len(raw_ss) >= 1:
                    self.start = raw_ss[0]
                    self.Log("No set <start>, Auto Set : {start <- %s}" % self.start)
                elif self.start is not None and self.end is None and len(raw_ss) >= 1:
                    self.end = raw_ss[-1]
                    self.Log("No set <end>, Auto Set : {end <- %s}" % self.end)
                else:
                    raise Exception("\n\n[Error]: StateMachine must have a start-node and a end-node.\n\t--At least, you need pass stateSet which length >= 2.")


    def Build(self):  # 编译
        """
        编译状态机
        在启用状态机前必须至少编译一次
        状态机运行时会执行最近一次的编译结果
        :return:
        """
        def cross_chain_iter(_tf):  # 转移函数展开 迭代器
            for _srt in _tf.srt:
                for _dst in _tf.dst:
                    yield _srt, _dst, _tf,   # srt, dst, tf

        # ------------- 子函数 区域 -------------------
        raw_ss, raw_tfs = self.GetTidiedInputParams()

        # 检查start和end
        self.TidyStartAndEnd(raw_ss, raw_tfs)

        self.state_table = {}
        self.st = self.state_table  # 别名

        # 生成keys for 'state table'
        if not self.raw_ss_auto_fill_flag:
            for sta in raw_ss:
                self.st[sta] = []  # name: []

        # 考虑优先级，分配这些函数
        for bundled_tf in raw_tfs:
            tf_iter = cross_chain_iter(bundled_tf)  # 获取传递函数迭代器(解除交叉参数)
            for srt, dst, tf in tf_iter:  # 对一次TransFunc所设定的所有link 按优先级排布
                added_flag = False   # 如果在排序中没有添加此项，那么应当在末尾添加此项
                if self.raw_ss_auto_fill_flag:
                    if srt not in self.st:
                        self.st[srt] = []
                        self.Log("[Auto] Add srt<%s> to state table." % srt)
                    if dst not in self.st:
                        self.st[dst] = []
                        self.Log("[Auto] Add dst<%s> to state table." % dst)
                else:
                    if self.st.get(srt) is None:
                        raise Exception("\n\n[Error]: This <TransFunc>'s srt: <{}> doesn't exist in state sets.\n\n{}".format(srt, "srt:\n\t{},\ndst:\n\t{}".format(bundled_tf.srt, bundled_tf.dst)))
                if self.st[srt]:  # 按priority进行排序，高优先级的在前
                    for index in range(len(self.st[srt])):
                        _srt, _dst, _tf = self.st[srt][index]
                        if tf.priority <= _tf.priority: continue
                        else:
                            self.st[srt].insert(index, (srt, dst, tf))
                            added_flag = True

                if not added_flag:  # 在排序中没有添加此项，那么应当在末尾添加此项
                    self.st[srt] += [(srt, dst, tf)]



        # 移除重复项
        for sta in self.st:
            self.st[sta] = list(set(self.st[sta]))

        # Build完成
        self.ready = self.READY

        self.ResetAreas()  # 重置各级运行环境


class StateMachine(StateMachine_cb, StateMachine_area, StateMachine_data):  # 状态机
    def __init__(self, *args, start=None, end=None):
        """
        用户可以只传trans_funcs这个参数，PyState会自动提取节点名
        :param arg1: 可以为state sets (记做sets tfs 正常模式)， 也可以是tfs (记做auto-sets tfs 自动补全模式)
        :param arg2: 正常模式时: tfs; 自动补全模式时: start(可省略)
        :param arg3: 正常模式时: start(可省略); 自动补全模式时: end(可省略)
        :param arg4: 正常模式时: end(可省略); 自动补全模式时: 不需要该参数，传入会产生歧义，可能会导致程序产生Exception

        当start和end通过kwargs传递时，可以忽略args中的start和end
        """
        StateMachine_data.__init__(self, *args, start=start, end=end)

        self.Loading()


    def Loading(self, *args, **kwargs):
        ...

    """
    def Build(self):  # 编译
        
    编译状态机
    在启用状态机前必须至少编译一次
    状态机运行时会执行最近一次的编译结果
    :return:
    """


    def Stop(self):
        """
        停止当前状态机，进入finish状态
        注意：停止后无法继续Execute。
            如果希望重新Execute，请在finish状态下调用Reset方法
        :return:
        """
        if self.ready < self.READY:
            raise NotImplementBuildUp()
        self.ready = self.FINISH
        if self.end_cb:
            self.end_cb(self)

    def __exec(self):
        """
        底层状态机执行函数
        :return:
        """
        ChangeGlobal_Machine(self.machine_area, self.machine_area)
        ChangeGlobal_Inst(self.machine_area, self.inst_areas[self.now])
        for i in range(len(self.st[self.now])):  # 按照优先级执行
            srt, dst, tf = self.st[self.now][i]
            ChangeGlobal_Dest(self.machine_area, self.inst_areas[dst])
            ChangeGlobal_This(self.machine_area, self.this_areas[self.now][i])
            sta = tf.Trig()

            if sta:  # 转移
                for _srt, _dst, _tf in self.st[self.now]: tf.buffer.Reset()  # 重置所有缓冲
                self.now = dst  # 状态转移
                self.log += [srt + " -> " + dst]
                if self.trans_cb:
                    self.trans_cb(self, srt, dst, tf)

                if self.now == self.end:
                    self.Log("FSM Touch End<%s>."%self.end)
                    self.Stop()
                break
        ChangeGlobalBack_NoCheck(self.machine_area)

    def Execute(self):
        """
        执行一步状态机
        :return:
        """
        if self.ready < self.READY:
            raise NotImplementBuildUp()
        elif self.ready >= self.FINISH:
            raise Exception("\n\n[Error]: you must .Reset() before restarting it.")
        if self.ready < self.BUSY:
            self.now = self.start
            self.ready = self.BUSY
            self.log = []
            if self.start_cb:
                self.start_cb(self)

        if self.ready == self.BUSY:
            self.__exec()

    def Reset(self):
        """
        重置结束标志符(finish)，以便可以重新开始执行

        :return:
        """
        if self.ready <= self.READY:
            raise NotImplementBuildUp()
        self.ready = self.READY

    def IsFinish(self):
        """
        是否执行完成
        执行到<end>节点或是用户手动调用Stop()都会进入Finish状态
        进入Finish状态后需要手动调用Reset()来重新使用该状态机
        :return:
        """
        return self.ready >= self.FINISH

    def IsBuildUp(self):
        """
        是否编译完成
        只要编译过一次，便始终为True
        :return:
        """
        return self.ready >= self.READY

    def IsRunning(self):
        """
        是否正在运行
        注意: 至少需要执行一次Execute才会进入Running状态
        :return:
        """
        return self.ready >= self.BUSY

    def GetNodeArea(self, name):
        if self.ready < self.READY:
            raise NotImplementBuildUp()
        get = self.inst_areas[name]
        if get is None:
            raise StateNotInSets(name)
        return get

    def StateGraph(self):
        if not self.ready:
            raise NotImplementBuildUp()
        try:
            from graphviz import Digraph
        except ImportError:
            raise ImportError("\n\n[Error]: Unable to import graphviz, please install them.")
        # ------------------- 子函数 --------------------
        # ------------------- 子函数 --------------------

        graph = Digraph(comment='State Table Graph', graph_attr={"labeljust":'l'})
        # build nodes first
        for k in self.st:
            if self.start == k: graph.node(k, k + "<In>", {'fontsize':'16', 'margin': '0.2'}, color='red', fontname="Microsoft YaHei")
            elif self.end == k: graph.node(k, k + "<Out>", {'fontsize':'16', 'margin': '0.2'}, color='blue', fontname="Microsoft YaHei")
            else: graph.node(k, k, {'fontsize':'20', 'margin': '0.2'}, fontname="Microsoft YaHei")
        # build links second
        for k in self.st:
            for srt, dst, tf in self.st[k]:
                # build doc
                doc = ""
                if tf.priority:
                    doc += "[{}]".format(tf.priority)
                if tf.buffer.Max():
                    doc += "<buf[%d]>: "%tf.buffer.Max() + tf.buffer.doc + '\n'
                if tf.cond_doc:
                    doc += "<if>: " + tf.cond_doc + '\n'
                if tf.exec_doc:
                    doc += "<do>: " + tf.exec_doc + '\n'
                # add link
                graph.edge(srt, dst, doc, {"labeldistance":"0.5", "decorate":"True", "labelfontsize":"20", "minlen":"4"}, fontname="Microsoft YaHei")
        return graph

    def ShowLog(self, if_show=False):
        """
        启动/停止 日志的控制台输出.
        :param if_show: 启动/停止
        :return:
        """
        self.show_log_flag = if_show

TF = TransFunc
SM = StateMachine



